/******************************************************************************
 * $Id$
 *
 * Project:  Arc/Info Binary Grid Driver
 * Purpose:  Implements GDAL interface to underlying library.
 * Author:   Frank Warmerdam, warmerdam@pobox.com
 *
 ******************************************************************************
 * Copyright (c) 1999, Frank Warmerdam
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 ****************************************************************************/

#include "gdal_pam.h"
#include "cpl_string.h"
#include "ogr_spatialref.h"
#include "gdal_rat.h"
#include "aigrid.h"
#include "avc.h"

CPL_CVSID("$Id$");

CPL_C_START
void	GDALRegister_AIGrid(void);
CPL_C_END

static CPLString OSR_GDS( char **papszNV, const char * pszField, 
                           const char *pszDefaultValue );


/************************************************************************/
/* ==================================================================== */
/*				                 AIGDataset		                		*/
/* ==================================================================== */
/************************************************************************/

class AIGRasterBand;

class CPL_DLL AIGDataset : public GDALPamDataset
{
    friend class AIGRasterBand;
    
    AIGInfo_t	*psInfo;

    char	**papszPrj;
    char	*pszProjection;
    GDALDataType eType;

    GDALColorTable *poCT;
    int         bHasReadRat;

    void        TranslateColorTable( const char * );

    void        ReadRAT();
    GDALRasterAttributeTable *poRAT;
     static AIGInfo_t *CreateLL( const char * pszCoverName,
        int nXSize, int nYSize, GDALDataType eType,char **papszParmList );

    CPLErr WriteProjection( const char * pszCoverName );

  public:
                AIGDataset();
                ~AIGDataset();

    static GDALDataset *Open( GDALOpenInfo * );
    static GDALDataset *Create( const char * pszFilename,
                                   int nXSize, int nYSize, int nBands,
                                   GDALDataType eType,
                                   char **papszParmList );

    virtual CPLErr GetGeoTransform( double * );
    virtual CPLErr SetGeoTransform( double * );
    virtual const char *GetProjectionRef(void);
    virtual CPLErr SetProjection( const char * );
    virtual char **GetFileList(void);
    virtual void FlushCache(void);
};

/************************************************************************/
/* ==================================================================== */
/*                            AIGRasterBand                             */
/* ==================================================================== */
/************************************************************************/

class AIGRasterBand : public GDALPamRasterBand

{
    friend class AIGDataset;
    double outputNoData;
    

  public:

                   AIGRasterBand( AIGDataset *, int );

    virtual CPLErr IReadBlock( int, int, void * );
    virtual CPLErr IWriteBlock( int, int, void * );
    virtual double GetMinimum( int *pbSuccess );
    virtual double GetMaximum( int *pbSuccess );
    virtual double GetNoDataValue( int *pbSuccess );
    virtual CPLErr SetNoDataValue( double );

    virtual GDALColorInterp GetColorInterpretation();
    virtual GDALColorTable *GetColorTable();
    virtual const GDALRasterAttributeTable *GetDefaultRAT();
};

/************************************************************************/
/*                           AIGRasterBand()                            */
/************************************************************************/

AIGRasterBand::AIGRasterBand( AIGDataset *poDS, int nBand )

{
    this->poDS = poDS;
    this->nBand = nBand;

    nBlockXSize = poDS->psInfo->nBlockXSize;
    nBlockYSize = poDS->psInfo->nBlockYSize;

    if( poDS->GetAccess() == GA_ReadOnly )
    {

        if( poDS->psInfo->nCellType == AIG_CELLTYPE_INT
            && poDS->psInfo->dfMin >= 0.0 && poDS->psInfo->dfMax <= 254.0 )
        {
            eDataType = GDT_Byte;
        }
        else if( poDS->psInfo->nCellType == AIG_CELLTYPE_INT
            && poDS->psInfo->dfMin >= -32767 && poDS->psInfo->dfMax <= 32767 )
        {
            eDataType = GDT_Int16;
        }
        else if( poDS->psInfo->nCellType == AIG_CELLTYPE_INT )
        {
            eDataType = GDT_Int32;
        }
        else
        {
            eDataType = GDT_Float32;
        }
    }
    else
    {
        eDataType = poDS->eType;

        //set outputNoData at the defaut no data for the type
        if( eDataType == GDT_Byte)
            outputNoData=255;
        else if( eDataType == GDT_Int16)
            outputNoData=-32767;
        else if( eDataType == GDT_Int32)
            outputNoData=ESRI_GRID_NO_DATA;
        else
            outputNoData=ESRI_GRID_FLOAT_NO_DATA;
    }
}

/************************************************************************/
/*                             IReadBlock()                             */
/************************************************************************/

CPLErr AIGRasterBand::IReadBlock( int nBlockXOff, int nBlockYOff,
                                  void * pImage )

{
    AIGDataset	*poODS = (AIGDataset *) poDS;
    GInt32	*panGridRaster;
    int		i;


    if( poODS->eAccess == GA_Update )
    {
        if( eDataType == GDT_Byte )
        {
            for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
                ((GByte *) pImage)[i] = (GByte)255;
        }
        else if( eDataType == GDT_Int16 )
        {
            for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
                ((GInt16 *) pImage)[i] = (GInt16)-32768;
        }
        else if( eDataType == GDT_Int32 )
        {
            for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
                ((GInt32 *) pImage)[i] = (GInt32)ESRI_GRID_NO_DATA;
        }
        else //Float32 
        {
            for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
                ((float *) pImage)[i] = (float)ESRI_GRID_FLOAT_NO_DATA;
        }
       
        return CE_None;
    }

    if( poODS->psInfo->nCellType == AIG_CELLTYPE_INT )
    {
        panGridRaster = (GInt32 *) VSIMalloc3(4,nBlockXSize,nBlockYSize);
        if( panGridRaster == NULL ||
            AIGReadTile( poODS->psInfo, nBlockXOff, nBlockYOff, panGridRaster )
            != CE_None )
        {
            CPLFree( panGridRaster );
            return CE_Failure;
        }

        if( eDataType == GDT_Byte )
        {
            for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
            {
                if( panGridRaster[i] == ESRI_GRID_NO_DATA )
                    ((GByte *) pImage)[i] = 255;
                else
                    ((GByte *) pImage)[i] = (GByte) panGridRaster[i];
            }
        }
        else if( eDataType == GDT_Int16 )
        {
            for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
            {
                if( panGridRaster[i] == ESRI_GRID_NO_DATA )
                    ((GInt16 *) pImage)[i] = -32768;
                else
                    ((GInt16 *) pImage)[i] = (GInt16) panGridRaster[i];
            }
        }
        else
        {
            for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
                ((GInt32 *) pImage)[i] = panGridRaster[i];
        }
        
        CPLFree( panGridRaster );

        return CE_None;
    }
    else
    {
        return AIGReadFloatTile( poODS->psInfo, nBlockXOff, nBlockYOff,
                                 (float *) pImage );
    }
}

/************************************************************************/
/*                          UpdateStatistic()                           */
/************************************************************************/

void UpdateStatistic(AIGInfo_t * psInfo, GInt32 * panData)
{
    
/* -------------------------------------------------------------------- */
/*      Compute statistics.                                             */
/* -------------------------------------------------------------------- */
    float	*pafData = (float *) panData;
    int nPixels = psInfo->nBlockXSize * psInfo->nBlockYSize;
    int bNoData=FALSE;
    double value=0;
    int i;

    for( i = 0; i < nPixels; i++ ) 
    {
        if( psInfo->nCellType == AIG_CELLTYPE_INT )
        {
            value = (double)panData[i];
            bNoData = (panData[i] == ESRI_GRID_NO_DATA);
        }
        else
        {
            value = (double)pafData[i];
            bNoData = (pafData[i] == ESRI_GRID_FLOAT_NO_DATA);
        }

        if(!bNoData)
        {
                
            if( value < psInfo->dfMin )
                psInfo->dfMin = value;
            
            if( value > psInfo->dfMax )
                psInfo->dfMax = value;
            
            psInfo->dfSumX+=value;
            psInfo->dfSumX2+=value*value;
            psInfo->nbValues++;
        }
    }
}

/************************************************************************/
/*                             IWriteBlock()                            */
/************************************************************************/


CPLErr AIGRasterBand::IWriteBlock( int nBlockXOff, int nBlockYOff,
                                  void * pImage )

{
    AIGDataset	*poODS = (AIGDataset *) poDS;
    int nPixels=0; 
    int i=0;
    float   *pafData = (float *) pImage;
    nPixels = poODS->psInfo->nBlockXSize * poODS->psInfo->nBlockYSize;
    CPLErr eErr = CE_None;
        
    if( poODS->psInfo->nCellType == AIG_CELLTYPE_INT )
    {
/* -------------------------------------------------------------------- */
/*      Perform integer pre processing and convert no data.				*/
/* -------------------------------------------------------------------- */

        GInt32 *panGridRaster = (GInt32 *) VSIMalloc3(4,nBlockXSize,nBlockYSize);
        if( panGridRaster == NULL )
            return CE_Failure;


        if( eDataType == GDT_Byte )
        {
            for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
            {
                //Byte and no data is always difficult to manage...
                if( ((GByte *) pImage)[i] == 255 || ((GByte *) pImage)[i] == outputNoData )
                    panGridRaster[i] = ESRI_GRID_NO_DATA;
                else
                    panGridRaster[i] = (GInt32) ((GByte *) pImage)[i];
            }
        }
        else if( eDataType == GDT_Int16 )
        {
            for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
            {
                if( ((GInt16 *) pImage)[i] == -32768 || ((GInt16 *) pImage)[i] == outputNoData )
                    panGridRaster[i] = ESRI_GRID_NO_DATA;
                else
                    panGridRaster[i] = (GInt32) ((GInt16 *) pImage)[i];
            }
        }
        else
        {
            for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
            {
				//Note ESRI no data is -2147483647
                if( ((GInt32 *) pImage)[i] == -2147483648 || ((GInt32 *) pImage)[i] == outputNoData)
                    panGridRaster[i] = ESRI_GRID_NO_DATA;
                else
                    panGridRaster[i] = (GInt32) ((GInt32 *) pImage)[i];
            }
        }

        UpdateStatistic(poODS->psInfo, panGridRaster);
/* -------------------------------------------------------------------- */
/*      Write block.                                   		            */
/* -------------------------------------------------------------------- */
        
        eErr = AIGWriteTile( poODS->psInfo, nBlockXOff, nBlockYOff,
                              panGridRaster );

        CPLFree( panGridRaster );
    }
    else
    {

/* -------------------------------------------------------------------- */
/*      Convert outputNoData to ESRI no data.                           */
/* -------------------------------------------------------------------- */
    
        for( i = 0; i < nPixels; i++ )
        {
            if( pafData[i] == outputNoData )
            {
                pafData[i] = ESRI_GRID_FLOAT_NO_DATA;
            }
        }

        UpdateStatistic(poODS->psInfo, (GInt32 *) pImage);
/* -------------------------------------------------------------------- */
/*      Write block.                                   		            */
/* -------------------------------------------------------------------- */

        eErr = AIGWriteTile( poODS->psInfo, nBlockXOff, nBlockYOff,
                                  (GInt32 *) pImage );
    }

    return eErr;
}


/************************************************************************/
/*                           GetDefaultRAT()                            */
/************************************************************************/

const GDALRasterAttributeTable *AIGRasterBand::GetDefaultRAT()

{
    AIGDataset	*poODS = (AIGDataset *) poDS;

/* -------------------------------------------------------------------- */
/*      Read info raster attribute table, if present.                   */
/* -------------------------------------------------------------------- */
    if (!poODS->bHasReadRat)
    {
        poODS->ReadRAT();
        poODS->bHasReadRat = TRUE;
    }

    if( poODS->poRAT )
        return poODS->poRAT;
    else
        return GDALPamRasterBand::GetDefaultRAT();
}

/************************************************************************/
/*                             GetMinimum()                             */
/************************************************************************/

double AIGRasterBand::GetMinimum( int *pbSuccess )

{
    AIGDataset	*poODS = (AIGDataset *) poDS;

    if( pbSuccess != NULL )
        *pbSuccess = TRUE;

    return poODS->psInfo->dfMin;
}

/************************************************************************/
/*                             GetMaximum()                             */
/************************************************************************/

double AIGRasterBand::GetMaximum( int *pbSuccess )

{
    AIGDataset	*poODS = (AIGDataset *) poDS;

    if( pbSuccess != NULL )
        *pbSuccess = TRUE;

    return poODS->psInfo->dfMax;
}

/************************************************************************/
/*                           GetNoDataValue()                           */
/************************************************************************/

double AIGRasterBand::GetNoDataValue( int *pbSuccess )

{ 
    if( pbSuccess != NULL )
        *pbSuccess = TRUE;

    if( poDS->GetAccess() == GA_ReadOnly )
    {
        if( eDataType == GDT_Float32 )
            return ESRI_GRID_FLOAT_NO_DATA;
        else if( eDataType == GDT_Int16 )
            return -32768;
        else if( eDataType == GDT_Byte )
            return 255;
        else
            return ESRI_GRID_NO_DATA;
    }
    else
    {
        return outputNoData;
    }
}

/************************************************************************/
/*                           SetNoDataValue()                           */
/************************************************************************/
CPLErr AIGRasterBand::SetNoDataValue( double noData)
{
    outputNoData=noData;
    return CE_None;
}

/************************************************************************/
/*                       GetColorInterpretation()                       */
/************************************************************************/

GDALColorInterp AIGRasterBand::GetColorInterpretation()

{
    AIGDataset	*poODS = (AIGDataset *) poDS;

    if( poODS->poCT != NULL )
        return GCI_PaletteIndex;
    else
        return GDALPamRasterBand::GetColorInterpretation();
}

/************************************************************************/
/*                           GetColorTable()                            */
/************************************************************************/

GDALColorTable *AIGRasterBand::GetColorTable()

{
    AIGDataset	*poODS = (AIGDataset *) poDS;

    if( poODS->poCT != NULL )
        return poODS->poCT;
    else
        return GDALPamRasterBand::GetColorTable();
}

/************************************************************************/
/* ==================================================================== */
/*                            AIGDataset                                */
/* ==================================================================== */
/************************************************************************/


/************************************************************************/
/*                            AIGDataset()                              */
/************************************************************************/

AIGDataset::AIGDataset()

{
    psInfo = NULL;
    papszPrj = NULL;
    pszProjection = CPLStrdup("");
    poCT = NULL;
    poRAT = NULL;
    bHasReadRat = FALSE;
}

/************************************************************************/
/*                           ~AIGDataset()                            */
/************************************************************************/

AIGDataset::~AIGDataset()

{
    FlushCache();
    
    if( psInfo != NULL )
    {
        
        if( eAccess == GA_Update )
        { 
            //save all other file
            AIGWriteHeader( psInfo->pszCoverName, psInfo );
            AIGWriteBlockIndex( psInfo->pszCoverName, psInfo);
            AIGWriteBounds( psInfo->pszCoverName, psInfo );
            AIGWriteStatistics( psInfo->pszCoverName, psInfo );
            WriteProjection( psInfo->pszCoverName );
        

            GByte header[100] = {0};
    
            int nBlockID=psInfo->pasTileInfo->nBlocks-1;
            int nLength = psInfo->pasTileInfo->panBlockOffset[nBlockID]+psInfo->pasTileInfo->panBlockSize[nBlockID]+ 2 + 100;
            GInt32 nValue = CPL_MSBWORD32(nLength /2);

            memcpy( header, "\x0\x0\x27\x0a\xff\xff\xfc\x14", 8);
            memcpy( header+24, &(nValue), 4 );
            VSIFSeekL( psInfo->pasTileInfo->fpGrid, 0, SEEK_SET );
            VSIFWriteL( (void*)header, 1, 100, psInfo->pasTileInfo->fpGrid );
        }

        AIGClose( psInfo );
        psInfo=NULL;
    }

    CPLFree( pszProjection );
    CSLDestroy( papszPrj );
    if( psInfo != NULL )
        AIGClose( psInfo );

    if( poCT != NULL )
        delete poCT;

    if( poRAT != NULL )
        delete poRAT;
}

/************************************************************************/
/*                            FlushCache()                              */
/************************************************************************/
void AIGDataset::FlushCache(void)
{
    GDALPamDataset::FlushCache();

    if( psInfo != NULL && eAccess == GA_Update )
    { 
        //save statistics
        psInfo->dfMean = psInfo->dfSumX/psInfo->nbValues;
        psInfo->dfStdDev = sqrt( __max(0, (psInfo->dfSumX2 - (psInfo->dfSumX*psInfo->dfSumX)/psInfo->nbValues))/(psInfo->nbValues-1));

        //write statistim tyo PAM section before flush it
        char pszValue[100]={0};
        sprintf(pszValue, "%f", psInfo->dfMin);
        SetMetadataItem( "STATISTICS_MINIMUM", pszValue);
        sprintf(pszValue, "%f", psInfo->dfMax);
        SetMetadataItem( "STATISTICS_MAXIMUM", pszValue);
        sprintf(pszValue, "%f", psInfo->dfMean);
        SetMetadataItem( "STATISTICS_MEAN", pszValue);
        sprintf(pszValue, "%f", psInfo->dfStdDev);
        SetMetadataItem( "STATISTICS_STDDEV", pszValue);
    }
}

/************************************************************************/
/*                            GetFileList()                             */
/************************************************************************/

char **AIGDataset::GetFileList()

{
    char **papszFileList = GDALPamDataset::GetFileList();

    // Add in all files in the cover directory.
    char **papszCoverFiles = VSIReadDir( GetDescription() );
    int i;

    for( i = 0; papszCoverFiles != NULL && papszCoverFiles[i] != NULL; i++ )
    {
        if( EQUAL(papszCoverFiles[i],".")
            || EQUAL(papszCoverFiles[i],"..") )
            continue;

        papszFileList = 
            CSLAddString( papszFileList, 
                          CPLFormFilename( GetDescription(), 
                                           papszCoverFiles[i], 
                                           NULL ) );
    }
    CSLDestroy(papszCoverFiles);
    
    return papszFileList;
}

/************************************************************************/
/*                              ReadRAT()                               */
/************************************************************************/

void AIGDataset::ReadRAT()

{
#ifndef OGR_ENABLED
#else
/* -------------------------------------------------------------------- */
/*      Check if we have an associated info directory.  If not          */
/*      return quietly.                                                 */
/* -------------------------------------------------------------------- */
    CPLString osInfoPath, osTableName;
    VSIStatBufL sStatBuf;

    osInfoPath = psInfo->pszCoverName;
    osInfoPath += "/../info";
    
    if( VSIStatL( osInfoPath, &sStatBuf ) != 0 )
    {
        CPLDebug( "AIG", "No associated info directory at: %s, skip RAT.",
                  osInfoPath.c_str() );
        return;
    }
    
    osInfoPath += "/";

/* -------------------------------------------------------------------- */
/*      Attempt to open the VAT table associated with this coverage.    */
/* -------------------------------------------------------------------- */
    osTableName = CPLGetFilename(psInfo->pszCoverName);
    osTableName += ".VAT";

    AVCBinFile *psFile = 
        AVCBinReadOpen( osInfoPath, osTableName,
                        AVCCoverTypeUnknown, AVCFileTABLE, NULL );

    CPLErrorReset();
    if( psFile == NULL )
        return;

    AVCTableDef *psTableDef = psFile->hdr.psTableDef;

/* -------------------------------------------------------------------- */
/*      Setup columns in corresponding RAT.                             */
/* -------------------------------------------------------------------- */
    int iField;

    poRAT = new GDALRasterAttributeTable();

    for( iField = 0; iField < psTableDef->numFields; iField++ )
    {
        AVCFieldInfo *psFDef = psTableDef->pasFieldDef + iField;
        GDALRATFieldUsage eFUsage = GFU_Generic;
        GDALRATFieldType eFType = GFT_String;

        CPLString osFName = psFDef->szName;
        osFName.Trim();

        if( EQUAL(osFName,"VALUE") )
            eFUsage = GFU_MinMax;
        else if( EQUAL(osFName,"COUNT") )
            eFUsage = GFU_PixelCount;
        
        if( psFDef->nType1 * 10 == AVC_FT_BININT )
            eFType = GFT_Integer;
        else if( psFDef->nType1 * 10 == AVC_FT_BINFLOAT )
            eFType = GFT_Real;

        poRAT->CreateColumn( osFName, eFType, eFUsage );
    }

/* -------------------------------------------------------------------- */
/*      Process all records into RAT.                                   */
/* -------------------------------------------------------------------- */
    AVCField *pasFields;
    int iRecord = 0;

    while( (pasFields = AVCBinReadNextTableRec(psFile)) != NULL )
    {
        iRecord++;

        for( iField = 0; iField < psTableDef->numFields; iField++ )
        {
            switch( psTableDef->pasFieldDef[iField].nType1 * 10 )
            {
              case AVC_FT_DATE:
              case AVC_FT_FIXINT:
              case AVC_FT_CHAR:
              case AVC_FT_FIXNUM:
              {
                  // XXX - I bet mloskot would like to see const_cast + static_cast :-)
                  const char* pszTmp = (const char*)(pasFields[iField].pszStr);
                  CPLString osStrValue( pszTmp );
                  poRAT->SetValue( iRecord-1, iField, osStrValue.Trim() );
              }
              break;

              case AVC_FT_BININT:
                if( psTableDef->pasFieldDef[iField].nSize == 4 )
                    poRAT->SetValue( iRecord-1, iField, 
                                     pasFields[iField].nInt32 );
                else
                    poRAT->SetValue( iRecord-1, iField, 
                                     pasFields[iField].nInt16 );
                break;

              case AVC_FT_BINFLOAT:
                if( psTableDef->pasFieldDef[iField].nSize == 4 )
                    poRAT->SetValue( iRecord-1, iField, 
                                     pasFields[iField].fFloat );
                else
                    poRAT->SetValue( iRecord-1, iField, 
                                     pasFields[iField].dDouble );
                break;
            }
        }
    }

/* -------------------------------------------------------------------- */
/*      Cleanup                                                         */
/* -------------------------------------------------------------------- */

    AVCBinReadClose( psFile );

    /* Workaround against #2447 and #3031, to avoid binding languages */
    /* not being able to open the dataset */
    CPLErrorReset();

#endif /* OGR_ENABLED */
}

/************************************************************************/
/*                                Open()                                */
/************************************************************************/

GDALDataset *AIGDataset::Open( GDALOpenInfo * poOpenInfo )

{
    AIGInfo_t	*psInfo;

/* -------------------------------------------------------------------- */
/*      If the pass name ends in .adf assume a file within the          */
/*      coverage has been selected, and strip that off the coverage     */
/*      name.                                                           */
/* -------------------------------------------------------------------- */
    CPLString osCoverName;

    osCoverName = poOpenInfo->pszFilename;
    if( osCoverName.size() > 4 
        && EQUAL(osCoverName.c_str()+osCoverName.size()-4,".adf") )
    {
        osCoverName = CPLGetDirname( poOpenInfo->pszFilename );
        if( osCoverName == "" )
            osCoverName = ".";
    }

/* -------------------------------------------------------------------- */
/*      Otherwise verify we were already given a directory.             */
/* -------------------------------------------------------------------- */
    else if( !poOpenInfo->bIsDirectory )
    {
        return NULL;
    }

/* -------------------------------------------------------------------- */
/*      Verify that a few of the "standard" files are available.        */
/* -------------------------------------------------------------------- */
    VSIStatBufL sStatBuf;
    CPLString osTestName;
    
    osTestName.Printf( "%s/hdr.adf", osCoverName.c_str() );
    if( VSIStatL( osTestName, &sStatBuf ) != 0 )

    {
        osTestName.Printf( "%s/HDR.ADF", osCoverName.c_str() );
        if( VSIStatL( osTestName, &sStatBuf ) != 0 )
            return NULL;
    }

/* -------------------------------------------------------------------- */
/*      Confirm we have at least one raster data file.  These can be    */
/*      sparse so we don't require particular ones to exists but if     */
/*      there are none this is likely not a grid.                       */
/* -------------------------------------------------------------------- */
    char **papszFileList = VSIReadDir( osCoverName );
    int iFile;
    int bGotOne = FALSE;

    if (papszFileList == NULL)
    {
        /* Useful when reading from /vsicurl/ on servers that don't */
        /* return a file list */
        /* such as /vsicurl/http://eros.usgs.gov/archive/nslrsda/GeoTowns/NLCD/89110458 */
        do
        {
            osTestName.Printf( "%s/W001001.ADF", osCoverName.c_str() );
            if( VSIStatL( osTestName, &sStatBuf ) == 0 )
            {
                bGotOne = TRUE;
                break;
            }

            osTestName.Printf( "%s/w001001.adf", osCoverName.c_str() );
            if( VSIStatL( osTestName, &sStatBuf ) == 0 )
            {
                bGotOne = TRUE;
                break;
            }
        } while(0);
    }

    for( iFile = 0; 
         papszFileList != NULL && papszFileList[iFile] != NULL && !bGotOne;
         iFile++ )
    {
        if( strlen(papszFileList[iFile]) != 11 )
            continue;

        // looking for something like w001001.adf or z001013.adf
        if( papszFileList[iFile][0] != 'w'
            && papszFileList[iFile][0] != 'W'
            && papszFileList[iFile][0] != 'z'
            && papszFileList[iFile][0] != 'Z' )
            continue;

        if( strncmp(papszFileList[iFile] + 1, "0010", 4) != 0 )
            continue;

        if( !EQUAL(papszFileList[iFile] + 7, ".adf") )
            continue;

        bGotOne = TRUE;
    }
    CSLDestroy( papszFileList );

    if( !bGotOne )
        return NULL;
    
/* -------------------------------------------------------------------- */
/*      Open the file.                                                  */
/* -------------------------------------------------------------------- */
    psInfo = AIGOpen( osCoverName.c_str(), "r" );
    
    if( psInfo == NULL )
    {
        CPLErrorReset();
        return NULL;
    }
    
/* -------------------------------------------------------------------- */
/*      Confirm the requested access is supported.                      */
/* -------------------------------------------------------------------- */
    if( poOpenInfo->eAccess == GA_Update )
    {
        AIGClose(psInfo);
        CPLError( CE_Failure, CPLE_NotSupported, 
                  "The AIG driver does not support update access to existing"
                  " datasets.\n" );
        return NULL;
    }
/* -------------------------------------------------------------------- */
/*      Create a corresponding GDALDataset.                             */
/* -------------------------------------------------------------------- */
    AIGDataset 	*poDS;

    poDS = new AIGDataset();

    poDS->psInfo = psInfo;

/* -------------------------------------------------------------------- */
/*      Try to read a color table (.clr).  It seems it is legal to      */
/*      have more than one so we just use the first one found.          */
/* -------------------------------------------------------------------- */
    char **papszFiles = CPLReadDir( psInfo->pszCoverName );
    CPLString osClrFilename;
    CPLString osCleanPath = CPLCleanTrailingSlash( psInfo->pszCoverName );
  
    // first check for any .clr in coverage dir.
    for( iFile = 0; papszFiles != NULL && papszFiles[iFile] != NULL; iFile++ )
    {
        if( !EQUAL(CPLGetExtension(papszFiles[iFile]),"clr") && !EQUAL(CPLGetExtension(papszFiles[iFile]),"CLR"))
            continue;
      
        osClrFilename = CPLFormFilename( psInfo->pszCoverName,
                                         papszFiles[iFile], NULL );
        break;
    }
  
    CSLDestroy( papszFiles );
  
    // Look in parent if we don't find a .clr in the coverage dir.
    if( strlen(osClrFilename) == 0 )
    {
        osTestName.Printf( "%s/../%s.clr",
                           psInfo->pszCoverName,
                           CPLGetFilename( osCleanPath ) );
      
        if( VSIStatL( osTestName, &sStatBuf ) != 0 )

    {
            osTestName.Printf( "%s/../%s.CLR",
                               psInfo->pszCoverName,
                               CPLGetFilename( osCleanPath ) );
      
            if( !VSIStatL( osTestName, &sStatBuf ) )
                osClrFilename = osTestName;
        }
        else
            osClrFilename = osTestName;
    }
    
  
    if( strlen(osClrFilename) > 0 )
        poDS->TranslateColorTable( osClrFilename );
  
/* -------------------------------------------------------------------- */
/*      Establish raster info.                                          */
/* -------------------------------------------------------------------- */
    poDS->nRasterXSize = psInfo->nPixels;
    poDS->nRasterYSize = psInfo->nLines;
    poDS->nBands = 1;

/* -------------------------------------------------------------------- */
/*      Create band information objects.                                */
/* -------------------------------------------------------------------- */
    poDS->SetBand( 1, new AIGRasterBand( poDS, 1 ) );
    

/* -------------------------------------------------------------------- */
/*	Try to read projection file.					*/
/* -------------------------------------------------------------------- */
    const char	*pszPrjFilename;

    pszPrjFilename = CPLFormCIFilename( psInfo->pszCoverName, "prj", "adf" );
    if( VSIStatL( pszPrjFilename, &sStatBuf ) == 0 )
    {
        OGRSpatialReference	oSRS;

        poDS->papszPrj = CSLLoad( pszPrjFilename );

        if( oSRS.importFromESRI( poDS->papszPrj ) == OGRERR_NONE )
        {
            // If geographic values are in seconds, we must transform. 
            // Is there a code for minutes too? 
            if( oSRS.IsGeographic() 
                && EQUAL(OSR_GDS( poDS->papszPrj, "Units", ""), "DS") )
            {
                psInfo->dfLLX /= 3600.0;
                psInfo->dfURY /= 3600.0;
                psInfo->dfCellSizeX /= 3600.0;
                psInfo->dfCellSizeY /= 3600.0;
            }

            CPLFree( poDS->pszProjection );
            poDS->pszProjection=NULL;
            oSRS.exportToWkt( &(poDS->pszProjection) );
            if( poDS->pszProjection==NULL)
                poDS->pszProjection = CPLStrdup("");
        }
    }

/* -------------------------------------------------------------------- */
/*      Initialize any PAM information.                                 */
/* -------------------------------------------------------------------- */
    poDS->SetDescription( psInfo->pszCoverName );
    poDS->TryLoadXML();

/* -------------------------------------------------------------------- */
/*      Open overviews.                                                 */
/* -------------------------------------------------------------------- */
    poDS->oOvManager.Initialize( poDS, psInfo->pszCoverName );

    return( poDS );
}

/************************************************************************/
/*                            CreateLL()                                */
/*                                                                      */
/*      Shared functionality between AIGridDataset::Create() and        */
/*      AIGridCreateCopy() for creating AIGrdi file based on a set of   */
/*      options and a configuration.                                    */
/************************************************************************/
AIGInfo_t *AIGDataset::CreateLL( const char * pszCoverName,
                              int nXSize, int nYSize, 
                              GDALDataType eType,
                              char **papszParmList )

{
    char	*pszHDRFilename=NULL;
    VSIStatBuf   sStat;
    int nbBlockPerRowUtil=0;
    int nbBlockPerColUtil=0;
    AIGInfo_t *psInfo=NULL;

    if( eType != GDT_Byte && eType != GDT_Int16 &&
        eType != GDT_Int32 && eType != GDT_Float32 )
    {
        
    
        CPLError( CE_Failure, CPLE_NotSupported, 
                  "The AIG driver support only GDT_Byte, GDT_Int16, GDT_Int32 and GDT_Float32 for writing\n" );
        return NULL;
    }

    
// -------------------------------------------------------------------- 
//      Does the target directory already exist?  If not try to         
//      create it.                                                      
// -------------------------------------------------------------------- 
    if( CPLStat( pszCoverName, &sStat ) != 0 )
    {
        if( VSIMkdir( pszCoverName, 0755 ) != 0 )
        {
            CPLError( CE_Failure, CPLE_OpenFailed,
                      "Unable to find, or create directory `%s'.",
                      pszCoverName );

            return NULL;
        }
    }

    psInfo = (AIGInfo_t *) CPLCalloc(sizeof(AIGInfo_t),1);

    if (psInfo == NULL)
    {
        CPLError(CE_Failure, CPLE_OutOfMemory, "Cannot allocate tile info array");
        return NULL;
    }
    
    psInfo->bHasWarned = FALSE; 
    psInfo->pszCoverName = CPLStrdup( pszCoverName );
    

// -------------------------------------------------------------------- 
//      Set the number of pixels and lines.                         
// -------------------------------------------------------------------- 
    psInfo->nPixels = nXSize;
    psInfo->nLines = nYSize;
    
    psInfo->nTilesPerRow = 1;
    psInfo->nTilesPerColumn = 1;
    
    psInfo->nBlockXSize = nXSize<(256*65535)?256:1024;
    psInfo->nBlockYSize = nYSize<(16*65535)?16:1024;
    psInfo->nBlocksPerRow = (int)ceil( (double)psInfo->nPixels/psInfo->nBlockXSize );
    psInfo->nBlocksPerColumn = (int)ceil( (double)psInfo->nLines/psInfo->nBlockYSize);
    
    psInfo->dfMin = DBL_MAX;
    psInfo->dfMax = -DBL_MAX;
    psInfo->dfMean = 0;
    psInfo->dfStdDev = 0;
    psInfo->dfSumX=0;
    psInfo->dfSumX2=0;
    psInfo->nbValues=0;
    
    nbBlockPerRowUtil = (int)ceil( (double)psInfo->nPixels/psInfo->nBlockXSize );
    nbBlockPerColUtil = (int)ceil( (double)psInfo->nLines/psInfo->nBlockYSize);


    //the real number of blocks is equal to 
    psInfo->pasTileInfo = (AIGTileInfo *) VSICalloc(sizeof(AIGTileInfo), 1);
    
                  
    if (psInfo->pasTileInfo == NULL)
    {
        CPLError(CE_Failure, CPLE_OutOfMemory, "Cannot allocate tile info array");
        AIGClose( psInfo );
        return NULL;
    }

    psInfo->pasTileInfo->nBlocks = psInfo->nBlocksPerRow*psInfo->nBlocksPerColumn;

// -------------------------------------------------------------------- 
//	Select cell type.				
// -------------------------------------------------------------------- 

    if( eType == GDT_Float32 )
    {
        psInfo->nCellType = AIG_CELLTYPE_FLOAT; //Write float grid
    }
    else
    {
        psInfo->nCellType = AIG_CELLTYPE_INT; //Write integer grid
        psInfo->bCompressed = TRUE; //by default, interger grid is compressed
    }

// -------------------------------------------------------------------- 
//	Allocate AIGInfo block info arrays.				
// -------------------------------------------------------------------- 

    psInfo->pasTileInfo->panBlockOffset = (GUInt32 *) CPLMalloc(4 * psInfo->pasTileInfo->nBlocks);
    psInfo->pasTileInfo->panBlockSize = (int *) CPLMalloc(4 * psInfo->pasTileInfo->nBlocks);
    memset( psInfo->pasTileInfo->panBlockOffset, psInfo->pasTileInfo->nBlocks, 4 );
    memset( psInfo->pasTileInfo->panBlockSize, psInfo->pasTileInfo->nBlocks, 4 );
    
// -------------------------------------------------------------------- 
//      Open the file w001001.adf file itself.                          
// -------------------------------------------------------------------- 

    pszHDRFilename = (char *) CPLMalloc(strlen(pszCoverName)+30);
    sprintf( pszHDRFilename, "%s/w001001.adf", pszCoverName );
    psInfo->pasTileInfo->fpGrid = VSIFOpenL( pszHDRFilename, "w+" );
    
    if( psInfo->pasTileInfo->fpGrid == NULL )
    {
        CPLError( CE_Failure, CPLE_OpenFailed,
                  "Failed to open data file: \n%s\n",
                  pszHDRFilename );

        AIGClose( psInfo );
        return NULL;
    
    }
    
    //Seek to the beggining of the data
    VSIFSeekL( psInfo->pasTileInfo->fpGrid, 100, SEEK_SET );

    return psInfo;
}


/************************************************************************/
/*                               Create()                               */
/*                                                                      */
/*      Create a new AIGrid file.										*/
/************************************************************************/

GDALDataset *AIGDataset::Create( const char * pszFilename,
                                   int nXSize, int nYSize, int nBands,
                                   GDALDataType eType,
                                   char **papszParmList )

{
    AIGDataset*	poDS;
    AIGInfo_t	*psInfo;

/* -------------------------------------------------------------------- */
/*      If the pass name ends in .adf assume a file within the          */
/*      coverage has been selected, and strip that off the coverage     */
/*      name.                                                           */
/* -------------------------------------------------------------------- */
    CPLString osCoverName;

    osCoverName = pszFilename;
    if( osCoverName.size() > 4 
        && EQUAL(osCoverName.c_str()+osCoverName.size()-4,".adf") )
    {
        osCoverName = CPLGetDirname( pszFilename );
        if( osCoverName == "" )
            osCoverName = ".";
    }

/* -------------------------------------------------------------------- */
/*      Create the aiGrid file.                                         */
/* -------------------------------------------------------------------- */
    psInfo = CreateLL( osCoverName.c_str(), nXSize, nYSize, 
                        eType, papszParmList );

    if( psInfo == NULL )
        return NULL;

/* -------------------------------------------------------------------- */
/*      Create the new AIGDataset object.                               */
/* -------------------------------------------------------------------- */
    poDS = new AIGDataset();
    poDS->psInfo = psInfo;
  
/* -------------------------------------------------------------------- */
/*      Establish raster info.                                          */
/* -------------------------------------------------------------------- */
    poDS->nRasterXSize = psInfo->nPixels;
    poDS->nRasterYSize = psInfo->nLines;
    poDS->nBands = 1;
    poDS->eAccess = GA_Update;
    poDS->eType = eType;

/* -------------------------------------------------------------------- */
/*      Create band information objects.                                */
/* -------------------------------------------------------------------- */
    poDS->SetBand( 1, new AIGRasterBand( poDS, 1 ) );


/* -------------------------------------------------------------------- */
/*      Initialize any PAM information.                                 */
/* -------------------------------------------------------------------- */
    poDS->PamInitialize();
    poDS->SetDescription( psInfo->pszCoverName );
    

/* -------------------------------------------------------------------- */
/*      Open overviews.                                                 */
/* -------------------------------------------------------------------- */
    poDS->oOvManager.Initialize( poDS, psInfo->pszCoverName );

    return( poDS );
}

// -------------------------------------------------------------------- 
//	Write projection file.					
// -------------------------------------------------------------------- 
CPLErr AIGDataset::WriteProjection( const char * pszCoverName )
{
    char	*pszPrjFilename = (char *) CPLMalloc(strlen(pszCoverName)+30);
    sprintf( pszPrjFilename, "%s/prj.adf", pszCoverName );

    VSILFILE	*fp = VSIFOpenL( pszPrjFilename, "wt" );
    if (fp != NULL)
    {
        OGRSpatialReference     oSRS;
        char *pszBufPtr = pszProjection;
        oSRS.importFromWkt( (char **) &pszBufPtr );
        oSRS.morphToESRI();
        
        char                    *pszESRIProjection = NULL;
        oSRS.exportToWkt( &pszESRIProjection );
        VSIFWriteL( pszESRIProjection, 1, strlen(pszESRIProjection), fp );

        VSIFCloseL( fp );
        CPLFree( pszESRIProjection );
    }
    else
    {
        CPLError( CE_Failure, CPLE_FileIO, 
                    "Unable to create file %s.\n", pszPrjFilename );
        return CE_Failure;
    }

    CPLFree( pszPrjFilename );

    return( CE_None );
}


/************************************************************************/
/*                          GetGeoTransform()                           */
/************************************************************************/

CPLErr AIGDataset::GetGeoTransform( double * padfTransform )

{
    padfTransform[0] = psInfo->dfLLX;
    padfTransform[1] = psInfo->dfCellSizeX;
    padfTransform[2] = 0;

    padfTransform[3] = psInfo->dfURY;
    padfTransform[4] = 0;
    padfTransform[5] = -psInfo->dfCellSizeY;
    
    return( CE_None );
}

/************************************************************************/
/*                          SetGeoTransform()                           */
/************************************************************************/

CPLErr AIGDataset::SetGeoTransform( double * padfTransform )

{
    if( GetAccess() == GA_Update )
    {
        psInfo->dfLLX = padfTransform[0];
        psInfo->dfCellSizeX = padfTransform[1];
        psInfo->dfURY = padfTransform[3];
        psInfo->dfCellSizeY = -padfTransform[5];
        
        psInfo->dfURX = psInfo->dfLLX + (psInfo->nPixels*psInfo->dfCellSizeX);// + 0.5 * psInfo->dfCellSizeX);
        psInfo->dfLLY = psInfo->dfURY - (psInfo->nLines*psInfo->dfCellSizeY);// + 0.5 * psInfo->dfCellSizeY);

        return( CE_None );
    }
    else
    {
        CPLError( CE_Failure, CPLE_NotSupported,
                  "Attempt to call SetGeoTransform() on a read-only AIGrid file." );
        return CE_Failure;
    }
}


/************************************************************************/
/*                          GetProjectionRef()                          */
/************************************************************************/

const char *AIGDataset::GetProjectionRef()

{
    return pszProjection;
}

/************************************************************************/
/*                           SetProjection()                            */
/************************************************************************/

CPLErr AIGDataset::SetProjection( const char * pszNewProjection )

{

    if( !EQUALN(pszNewProjection,"GEOGCS",6)
        && !EQUALN(pszNewProjection,"PROJCS",6)
        && !EQUALN(pszNewProjection,"LOCAL_CS",8)
        && !EQUALN(pszNewProjection,"COMPD_CS",8)
        && !EQUAL(pszNewProjection,"") )
    {
        CPLError( CE_Failure, CPLE_AppDefined,
                "Only OGC WKT Projections supported for writing to AIGrid.\n"
                "%s not supported.",
                  pszNewProjection );
        
        return CE_Failure;
    }
    
    CPLFree( pszProjection );
    pszProjection = CPLStrdup( pszNewProjection );


    return CE_None;
}


/************************************************************************/
/*                        TranslateColorTable()                         */
/************************************************************************/

void AIGDataset::TranslateColorTable( const char *pszClrFilename )

{
    int  iLine;
    char **papszClrLines;

    papszClrLines = CSLLoad( pszClrFilename );
    if( papszClrLines == NULL )
        return;

    poCT = new GDALColorTable();

    for( iLine = 0; papszClrLines[iLine] != NULL; iLine++ )
    {
        char **papszTokens = CSLTokenizeString( papszClrLines[iLine] );

        if( CSLCount(papszTokens) >= 4 && papszTokens[0][0] != '#' )
        {
            int nIndex;
            GDALColorEntry sEntry;

            nIndex = atoi(papszTokens[0]);
            sEntry.c1 = (short) atoi(papszTokens[1]);
            sEntry.c2 = (short) atoi(papszTokens[2]);
            sEntry.c3 = (short) atoi(papszTokens[3]);
            sEntry.c4 = 255;

            if( (nIndex < 0 || nIndex > 33000) 
                || (sEntry.c1 < 0 || sEntry.c1 > 255)
                || (sEntry.c2 < 0 || sEntry.c2 > 255)
                || (sEntry.c3 < 0 || sEntry.c3 > 255) )
            {
                CSLDestroy( papszTokens );
                CPLError( CE_Failure, CPLE_AppDefined, 
                          "Color table entry appears to be corrupt, skipping the rest. " );
                break;
            }

            poCT->SetColorEntry( nIndex, &sEntry );
        }

        CSLDestroy( papszTokens );
    }

    CSLDestroy( papszClrLines );
}

/************************************************************************/
/*                              OSR_GDS()                               */
/************************************************************************/

static CPLString OSR_GDS( char **papszNV, const char * pszField, 
                           const char *pszDefaultValue )

{
    int         iLine;

    if( papszNV == NULL || papszNV[0] == NULL )
        return pszDefaultValue;

    for( iLine = 0; 
         papszNV[iLine] != NULL && 
             !EQUALN(papszNV[iLine],pszField,strlen(pszField));
         iLine++ ) {}

    if( papszNV[iLine] == NULL )
        return pszDefaultValue;
    else
    {
        CPLString osResult;
        char    **papszTokens;
        
        papszTokens = CSLTokenizeString(papszNV[iLine]);

        if( CSLCount(papszTokens) > 1 )
            osResult = papszTokens[1];
        else
            osResult = pszDefaultValue;
        
        CSLDestroy( papszTokens );
        return osResult;
    }
}

/************************************************************************/
/*                             AIGRename()                              */
/*                                                                      */
/*      Custom renamer for AIG dataset.                                 */
/************************************************************************/

static CPLErr AIGRename( const char *pszNewName, const char *pszOldName )

{
/* -------------------------------------------------------------------- */
/*      Make sure we are talking about paths to the coverage            */
/*      directory.                                                      */
/* -------------------------------------------------------------------- */
    CPLString osOldPath, osNewPath;

    if( strlen(CPLGetExtension(pszNewName)) > 0 )
        osNewPath = CPLGetPath(pszNewName);
    else
        osNewPath = pszNewName;

    if( strlen(CPLGetExtension(pszOldName)) > 0 )
        osOldPath = CPLGetPath(pszOldName);
    else
        osOldPath = pszOldName;

/* -------------------------------------------------------------------- */
/*      Get file list.                                                  */
/* -------------------------------------------------------------------- */

    GDALDatasetH hDS = GDALOpen( osOldPath, GA_ReadOnly );
    if( hDS == NULL )
        return CE_Failure;

    char **papszFileList = GDALGetFileList( hDS );
    GDALClose( hDS );

    if( papszFileList == NULL )
        return CE_Failure;

/* -------------------------------------------------------------------- */
/*      Work out the corresponding new names.                           */
/* -------------------------------------------------------------------- */
    char **papszNewFileList = NULL;
    int i;
    
    for( i = 0; papszFileList[i] != NULL; i++ )
    {
        CPLString osNewFilename;

        if( !EQUALN(papszFileList[i],osOldPath,strlen(osOldPath)) )
        {
            CPLAssert( FALSE );
            return CE_Failure;
        }

        osNewFilename = osNewPath + (papszFileList[i] + strlen(osOldPath));
        
        papszNewFileList = CSLAddString( papszNewFileList, osNewFilename );
    }

/* -------------------------------------------------------------------- */
/*      Try renaming the directory.                                     */
/* -------------------------------------------------------------------- */
    if( VSIRename( osNewPath, osOldPath ) != 0 )
    {
        if( VSIMkdir( osNewPath, 0777 ) != 0 )
        {
            CPLError( CE_Failure, CPLE_AppDefined, 
                      "Unable to create directory %s:\n%s",
                      osNewPath.c_str(),
                      VSIStrerror(errno) );
            return CE_Failure;
        }
    }

/* -------------------------------------------------------------------- */
/*      Copy/rename any remaining files.                                */
/* -------------------------------------------------------------------- */
    VSIStatBufL sStatBuf;

    for( i = 0; papszFileList[i] != NULL; i++ )
    {
        if( VSIStatL( papszFileList[i], &sStatBuf ) == 0 
            && VSI_ISREG( sStatBuf.st_mode ) )
        {
            if( CPLMoveFile( papszNewFileList[i], papszFileList[i] ) != 0 )
            {
                CPLError( CE_Failure, CPLE_AppDefined, 
                          "Unable to move %s to %s:\n%s",
                          papszFileList[i],
                          papszNewFileList[i], 
                          VSIStrerror(errno) );
                return CE_Failure;
            }
        }
    }

    if( VSIStatL( osOldPath, &sStatBuf ) == 0 )
        CPLUnlinkTree( osOldPath );

    return CE_None;
}

/************************************************************************/
/*                             AIGDelete()                              */
/*                                                                      */
/*      Custom dataset deleter for AIG dataset.                         */
/************************************************************************/

static CPLErr AIGDelete( const char *pszDatasetname )

{
/* -------------------------------------------------------------------- */
/*      Get file list.                                                  */
/* -------------------------------------------------------------------- */
    GDALDatasetH hDS = GDALOpen( pszDatasetname, GA_ReadOnly );
    if( hDS == NULL )
        return CE_Failure;

    char **papszFileList = GDALGetFileList( hDS );
    GDALClose( hDS );

    if( papszFileList == NULL )
        return CE_Failure;

/* -------------------------------------------------------------------- */
/*      Delete all regular files.                                       */
/* -------------------------------------------------------------------- */
    int i;
    for( i = 0; papszFileList[i] != NULL; i++ )
    {
        VSIStatBufL sStatBuf;
        if( VSIStatL( papszFileList[i], &sStatBuf ) == 0 
            && VSI_ISREG( sStatBuf.st_mode ) )
        {
            if( VSIUnlink( papszFileList[i] ) != 0 )
            {
                CPLError( CE_Failure, CPLE_AppDefined, 
                          "Unable to delete '%s':\n%s", 
                          papszFileList[i], VSIStrerror( errno ) );
                return CE_Failure;
            }
        }
    }

/* -------------------------------------------------------------------- */
/*      Delete directories.                                             */
/* -------------------------------------------------------------------- */
    for( i = 0; papszFileList[i] != NULL; i++ )
    {
        VSIStatBufL sStatBuf;
        if( VSIStatL( papszFileList[i], &sStatBuf ) == 0 
            && VSI_ISDIR( sStatBuf.st_mode ) )
        {
            if( CPLUnlinkTree( papszFileList[i] ) != 0 )
                return CE_Failure;
        }
    }

    return CE_None;
}

/************************************************************************/
/*                          GDALRegister_AIG()                          */
/************************************************************************/

void GDALRegister_AIGrid()

{
    GDALDriver	*poDriver;

    if( GDALGetDriverByName( "AIG" ) == NULL )
    {
        poDriver = new GDALDriver();
        
        poDriver->SetDescription( "AIG" );
        poDriver->SetMetadataItem( GDAL_DMD_LONGNAME, 
                                   "Arc/Info Binary Grid" ); 
        poDriver->SetMetadataItem( GDAL_DMD_HELPTOPIC, 
                                   "frmt_various.html#AIG" );
        poDriver->SetMetadataItem( GDAL_DCAP_VIRTUALIO, "YES" );
        
        poDriver->pfnOpen = AIGDataset::Open;
        poDriver->pfnCreate = AIGDataset::Create;

        poDriver->pfnRename = AIGRename;
        poDriver->pfnDelete = AIGDelete;

        GetGDALDriverManager()->RegisterDriver( poDriver );
    }
}
