/******************************************************************************
 * $Id$
 *
 * Project:  Arc/Info Binary Grid Translator
 * Purpose:  Grid file writing code.
 * Author:   Remi Saint-Amant from Frank Warmerdam code
 *
 ******************************************************************************
 * Copyright (c) 1999, Frank Warmerdam
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 ****************************************************************************/

#include "aigrid.h"
#include <math.h>
#include "cpl_string.h"


CPL_CVSID("$Id$");


static 
int AIGProcessRaw8BitBlockW( GInt32 * panData, int nMin, 
                                int nBlockXSize, int nBlockYSize, 
                                GByte *pabyCur);

static 
int AIGProcessRaw16BitBlockW( GInt32 * panData, int nMin, 
                                 int nBlockXSize, int nBlockYSize, 
                                 GByte *pabyCur);

static 
int AIGProcessRaw32BitBlockW( GInt32 * panData, GInt32 nMin, 
                                 int nBlockXSize, int nBlockYSize,
                                 GByte *pabyCur );

static 
int AIGProcessRaw32BitFloatBlockW( float * pafData, 
                                      int nBlockXSize, int nBlockYSize,
                                      GByte *pabyCur);
static 
int AIGProcessBlockW( GInt32 * panData, int nMin, int nMagic, 
                        int nBlockXSize, int nBlockYSize,  GByte *pabyCur);


CPLErr AIGWriteBlock( VSILFILE * fp, GUInt32* pBlockOffset, int* pBlockSize,
                      int nBlockXSize, int nBlockYSize,
                      GInt32 *panData, int nCellType, int bCompressed )
{
    int nBlockSize=0;
    GByte	*pabyRaw=NULL;
    int nBlockMaxSize = nBlockXSize*nBlockYSize*4;

    CPLAssert( nCellType == AIG_CELLTYPE_INT || nCellType == AIG_CELLTYPE_FLOAT );

    pabyRaw = (GByte *) CPLMalloc(nBlockMaxSize+2);
    
    if( nCellType == AIG_CELLTYPE_INT && bCompressed)
    {
// --------------------------------------------------------------------
//      Integer process.                                   
// --------------------------------------------------------------------
        int bNoData=FALSE;
        int nPixel = nBlockXSize*nBlockYSize;
        int nMin=INT_MAX;
        int nMax=INT_MIN;    
        int nMinMSB=0;
        int i=0;
        int dataSize=0;
        int nBite=0;

// --------------------------------------------------------------------
//      Find min, max and noData.                                   
// --------------------------------------------------------------------

        for( i = 0; i < nPixel; i++)
        {
            if( panData[i] != ESRI_GRID_NO_DATA )
            {
                nMin = __min( nMin, panData[i] );
                nMax = __max( nMax, panData[i] );
            }
            else
            {
                bNoData=TRUE;
            }
        }

        if( nMax-nMin <= 256)
            nBite=1;
        else if( nMax-nMin <= 65536)
            nBite=2;
        else
            nBite=4;

        CPLAssert( nBite== 1 || nBite==2 || nBite==4 );
/* -------------------------------------------------------------------- */
/*      Write data to buffer.                                           */
/* -------------------------------------------------------------------- */
        if( nMin==INT_MAX && nMax==INT_MIN ) 
        {
            nMin=ESRI_GRID_NO_DATA;
            pabyRaw[2] = 0x00;
            dataSize = 0;
        }
        else if( nMin==nMax && !bNoData )
        {
            pabyRaw[2] = 0x00;
            dataSize = 0;

        }
        else if( bNoData && nBite!=4 )
        {
            CPLAssert(nBite==1 || nBite==2);
            pabyRaw[2] = nBite==1?0xD7:0xCF;

            dataSize = AIGProcessBlockW( panData, nMin, pabyRaw[2], nBlockXSize, nBlockYSize, pabyRaw+8 );
        }
        else
        {
            pabyRaw[2] = (GByte)(8*nBite);

            if( nBite==1 )//8bits
                dataSize = AIGProcessRaw8BitBlockW( panData, nMin, nBlockXSize, nBlockYSize, pabyRaw+8 );
            else if( nBite==2 )//16bits
                dataSize = AIGProcessRaw16BitBlockW( panData, nMin, nBlockXSize, nBlockYSize, pabyRaw+8 );
            else if( nBite==4 )//32bits
                dataSize = AIGProcessRaw32BitBlockW( panData, nMin, nBlockXSize, nBlockYSize, pabyRaw+8 );
        }

        nBlockSize = dataSize+1+1+4;

        //BlockSize must be even
        if( nBlockSize%2 == 1)
            nBlockSize++;

        CPLAssert( nBlockSize < nBlockMaxSize );
        //write minimum value size
        pabyRaw[3] = 0x04;

        //write minimum value 
        nMinMSB = CPL_MSBWORD32( nMin );
        memcpy( &(pabyRaw[4]), &nMinMSB, 4 );

    }
    else
    {
/* -------------------------------------------------------------------- */
/*      Write data to buffer.                                           */
/* -------------------------------------------------------------------- */
        if( nCellType == AIG_CELLTYPE_INT )
            nBlockSize = AIGProcessRaw32BitBlockW( panData, 0, nBlockXSize, nBlockYSize, pabyRaw+2 );
        else
            nBlockSize = AIGProcessRaw32BitFloatBlockW( (float *)panData, nBlockXSize, nBlockYSize, pabyRaw+2);
    }
    

// -------------------------------------------------------------------- 
//      Writing block size.                                          
// -------------------------------------------------------------------- 
    pabyRaw[0] = (GByte)((nBlockSize/2)/256);
    pabyRaw[1] = (GByte)((nBlockSize/2)%256);

// -------------------------------------------------------------------- 
//      Update tile info table (offset and size).                      
// -------------------------------------------------------------------- 

    *pBlockOffset = (GUInt32)VSIFTellL( fp );
    *pBlockSize = nBlockSize;
    

// -------------------------------------------------------------------- 
//      Verify the block size.                                          
// -------------------------------------------------------------------- 
    CPLAssert( *pBlockSize == (pabyRaw[0]*256 + pabyRaw[1])*2 );

// --------------------------------------------------------------------
//      Write the block on disk.                                    
//-------------------------------------------------------------------- 
    VSIFWriteL( pabyRaw, *pBlockSize+2, 1, fp );
        
    CPLFree( pabyRaw );
    pabyRaw=NULL;

    return( CE_None );
}


//************************************************************************
//*                           AIGWriteHeader()                           
//*                                                                      
//*      Write the hdr.adf file, and populate the given info structure   
//*      appropriately.                                                  
//************************************************************************
static const unsigned char UNKNOWN_ZONE1_INT[8] = {0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF};
static const unsigned char UNKNOWN_ZONE1_FLOAT[8] = {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x20};
//the fisrt byte of UNKNOWN_ZONE2 seem to be 0x3F for float 
//and 0x3F, 0x40 or 0x41 for interger...
static const unsigned char UNKNOWN_ZONE2_INT[2] = {0x00,0x00};
static const unsigned char UNKNOWN_ZONE2_FLOAT[2] = {0x3F,0x80};



CPLErr AIGWriteHeader( const char * pszCoverName, AIGInfo_t * psInfo )
{
    GByte	abyData[308] = {0};
    char	*pszHDRFilename=NULL;
    FILE	*fp = NULL;
    GInt32 ONE = 1;
    GInt32 nCellType = psInfo->nCellType;
    GInt32 nUnCompressed = psInfo->bCompressed?0:1;
    GInt32 nBlocksPerRow = psInfo->nBlocksPerRow;
    GInt32 nBlocksPerColumn = psInfo->nBlocksPerColumn;
    GInt32 nBlockXSize = psInfo->nBlockXSize;
    GInt32 nBlockYSize = psInfo->nBlockYSize;
    double dfCellSizeX = psInfo->dfCellSizeX;
    double dfCellSizeY = psInfo->dfCellSizeY;
    double XTileRef = psInfo->dfLLX-(psInfo->nBlocksPerRow*psInfo->nBlockXSize*psInfo->dfCellSizeX)/2.0;
    double YTileRef = psInfo->dfURY-(3*psInfo->nBlocksPerColumn*psInfo->nBlockYSize*psInfo->dfCellSizeY)/2.0;
    const GByte *pUnknownZone1 = psInfo->nCellType==AIG_CELLTYPE_INT?UNKNOWN_ZONE1_INT:UNKNOWN_ZONE1_FLOAT;
    const GByte *pUnknownZone2 = psInfo->nCellType==AIG_CELLTYPE_INT?UNKNOWN_ZONE2_INT:UNKNOWN_ZONE2_FLOAT;
// -------------------------------------------------------------------- 
//      Open hdr.adf file.                                     
// -------------------------------------------------------------------- 
    pszHDRFilename = (char *) CPLMalloc(strlen(pszCoverName)+30);
    sprintf( pszHDRFilename, "%s/hdr.adf", pszCoverName );

    fp = VSIFOpen( pszHDRFilename, "wb" );
    
    if( fp == NULL )
    {
        CPLError( CE_Failure, CPLE_OpenFailed, 
                      "Unable to open header file for writing:\n%s\n",
                      pszHDRFilename);
        
        CPLFree( pszHDRFilename );
        return CE_Failure;
    }

    nCellType = CPL_MSBWORD32( nCellType );
    nUnCompressed = CPL_MSBWORD32( nUnCompressed );
    nBlocksPerRow = CPL_MSBWORD32( nBlocksPerRow );
    nBlocksPerColumn = CPL_MSBWORD32( nBlocksPerColumn );
    nBlockXSize = CPL_MSBWORD32( nBlockXSize );
    nBlockYSize = CPL_MSBWORD32( nBlockYSize );
    ONE = CPL_MSBWORD32( ONE );
    CPL_MSBPTR64( &(dfCellSizeX) );
    CPL_MSBPTR64( &(dfCellSizeY) );
    CPL_MSBPTR64( &(XTileRef) );
    CPL_MSBPTR64( &(YTileRef) );

   
// -------------------------------------------------------------------- 
//      Write the block size information.                                
//      There are some blanck
// -------------------------------------------------------------------- 
    memcpy( abyData, "GRID1.2\0", 8 );
    memcpy( abyData+8, pUnknownZone1, 8 );
    memcpy( abyData+16, &(nCellType), 4 );
    memcpy( abyData+20, &nUnCompressed, 4 );
    memcpy( abyData+28, pUnknownZone2, 2 );
    memcpy( abyData+256, &(dfCellSizeX), 8 );
    memcpy( abyData+264, &(dfCellSizeY), 8 );
    memcpy( abyData+272, &(XTileRef), 8 );
    memcpy( abyData+280, &(YTileRef), 8 );
    memcpy( abyData+288, &(nBlocksPerRow), 4 );
    memcpy( abyData+292, &(nBlocksPerColumn), 4 );
    memcpy( abyData+296, &(nBlockXSize), 4 );
    memcpy( abyData+300, &(ONE), 4 );
    memcpy( abyData+304, &(nBlockYSize), 4 );
    
    

// -------------------------------------------------------------------- 
//      Write the whole file (we expect it to always be 308 bytes        
//      long.                                                           
// -------------------------------------------------------------------- 

    VSIFWrite( abyData, 1, 308, fp );

    VSIFClose( fp );

    CPLFree( pszHDRFilename );
    return( CE_None );
}

//***********************************************************************
//*                         AIGWriteBlockIndex()                         
//*                                                                      
//*      Write the w001001x.adf file, and populate the given info        
//*      structure with the block offsets, and sizes.                    
//***********************************************************************
CPLErr AIGWriteBlockIndex( const char *pszCoverName, AIGInfo_t * psInfo)
{
    FILE	*fp=NULL;
    int		nLength=0, i=0;
    GUInt32	nValue=0;
    GUInt32	*panIndex=NULL;
    GByte firstByte[100] = {0};
    char	*pszHDRFilename=NULL;

// -------------------------------------------------------------------- 
//      Open w001001x.adf file.                                     
// -------------------------------------------------------------------- 
    pszHDRFilename = (char *) CPLMalloc(strlen(pszCoverName)+30);
    sprintf( pszHDRFilename, "%s/w001001x.adf", pszCoverName );

    fp = VSIFOpen( pszHDRFilename, "wb" );
    
    if( fp == NULL )
    {
        CPLError( CE_Failure, CPLE_OpenFailed, 
                    "Unable to open index blocks file for writing :\n%s\n",
                    pszHDRFilename);
        
        CPLFree( pszHDRFilename );
        return CE_Failure;
    }

    nLength = psInfo->pasTileInfo->nBlocks*8 + 100;
    nValue = CPL_MSBWORD32(nLength /2);

    memcpy( firstByte, "\x0\x0\x27\x0a\xff\xff\xfc\x14", 8);
    memcpy( firstByte+24, &(nValue), 4 );
    
// -------------------------------------------------------------------- 
//      Write the file length (in 2 byte shorts)                          
// -------------------------------------------------------------------- 
    VSIFWrite( &firstByte, 1, 100, fp );

    

// -------------------------------------------------------------------- 
//      Allocate buffer, and write the file 
//      from the buffer.                                                
// -------------------------------------------------------------------- 
    panIndex = (GUInt32 *) CPLMalloc(psInfo->pasTileInfo->nBlocks * 8);

// -------------------------------------------------------------------- 
//      Set block information.                                 
// -------------------------------------------------------------------- 
    CPLAssert(psInfo->pasTileInfo->panBlockOffset);
    CPLAssert(psInfo->pasTileInfo->panBlockSize);

    for( i = 0; i < psInfo->pasTileInfo->nBlocks; i++ )
    {
          panIndex[i*2] = CPL_MSBWORD32(psInfo->pasTileInfo->panBlockOffset[i]/2);
          panIndex[i*2+1] = CPL_MSBWORD32(psInfo->pasTileInfo->panBlockSize[i]/2);
    }

    VSIFWrite( panIndex, 8, psInfo->pasTileInfo->nBlocks, fp );

    VSIFClose( fp );
    
    CPLFree( panIndex );
    CPLFree( pszHDRFilename );

    return ( CE_None );
}

//***********************************************************************
//*                           AIGWriteBounds()                            
//*                                                                      
//*      Write the dblbnd.adf file for the georeferenced bounds.          
//***********************************************************************

CPLErr AIGWriteBounds( const char * pszCoverName, AIGInfo_t * psInfo )
{
    FILE	*fp=NULL;
    double	adfBound[4]={0};
    char	*pszHDRFilename=NULL;
// -------------------------------------------------------------------- 
//      Open the file dblbnd.adf file.                                  
// -------------------------------------------------------------------- 
    pszHDRFilename = (char *) CPLMalloc(strlen(pszCoverName)+30);
    sprintf( pszHDRFilename, "%s/dblbnd.adf", pszCoverName );

    fp = VSIFOpen( pszHDRFilename, "wb" );
    
    if( fp == NULL )
    {
        CPLError( CE_Failure, CPLE_OpenFailed,
                  "Failed to open grid bounds file: \n%s\n",
                  pszHDRFilename );

        CPLFree( pszHDRFilename );
        return( CE_Failure );
    }


    adfBound[0] = psInfo->dfLLX;
    adfBound[1] = psInfo->dfLLY;
    adfBound[2] = psInfo->dfURX;
    adfBound[3] = psInfo->dfURY;

    CPL_MSBPTR64(adfBound+0);
    CPL_MSBPTR64(adfBound+1);
    CPL_MSBPTR64(adfBound+2);
    CPL_MSBPTR64(adfBound+3);
    

// -------------------------------------------------------------------- 
//      Set the contents - four doubles.                                
// -------------------------------------------------------------------- 
    VSIFWrite( adfBound, 1, 32, fp );

    VSIFClose( fp );
    CPLFree( pszHDRFilename );

    return ( CE_None );
}

//***********************************************************************
//*                         AIGWriteStatistics()                          
//*                                                                      
//*      Write the sta.adf file for the layer statistics.                 
//***********************************************************************
CPLErr AIGWriteStatistics( const char * pszCoverName, AIGInfo_t * psInfo )
{
    FILE	*fp=NULL;
    double	adfStats[4]={0};
    char	*pszHDRFilename=NULL;
   
// -------------------------------------------------------------------- 
//      Open the file sta.adf file.                                     
// -------------------------------------------------------------------- 
    pszHDRFilename = (char *) CPLMalloc(strlen(pszCoverName)+30);
    sprintf( pszHDRFilename, "%s/sta.adf", pszCoverName );

    fp = VSIFOpen( pszHDRFilename, "wb" );
    
    if( fp == NULL )
    {
        CPLError( CE_Failure, CPLE_OpenFailed,
                  "Failed to open statistics file: \n%s\n",
                  pszHDRFilename );

        CPLFree( pszHDRFilename );
        return( CE_Failure );
    }

    adfStats[0] = psInfo->dfMin;
    adfStats[1] = psInfo->dfMax;
    adfStats[2] = psInfo->dfMean;
    adfStats[3] = psInfo->dfStdDev;

    CPL_MSBPTR64(adfStats+0);
    CPL_MSBPTR64(adfStats+1);
    CPL_MSBPTR64(adfStats+2);
    CPL_MSBPTR64(adfStats+3);

// -------------------------------------------------------------------- 
//      Set the contents - four doubles.                                
// -------------------------------------------------------------------- 
    VSIFWrite( adfStats, 1, 32, fp );

    VSIFClose( fp );

    return ( CE_None );
}


//CPLErr AIGWriteAUX(const char * pszCoverName, AIGInfo_t * psInfo )
//{
//	HFAHandle hHFA = NULL;
//	char	*pszHDRFilename=NULL;
//	//CString pszAUXFilePath = pszCoverName;
//    //pszAUXFilePath += ".aux";
//	pszHDRFilename = (char *) CPLMalloc(strlen(pszCoverName)+30);
//    sprintf( pszHDRFilename, "%s.aux", pszCoverName );
//
//
//	
//	char* papszOptions[1] = {NULL};
//	
//	hHFA = HFACreate( osHDRFilename.c_str(), psInfo->nPixels, psInfo->nLines, 
//							1, EPT_f32, psInfo->nBlockXSize, psInfo->nBlockYSize, papszOptions );
//	if( hHFA == NULL)
//        CPLError( CE_Failure, CPLE_OpenFailed,
//                  "Failed to open aux file: \n%s\n",
//                  osHDRFilename );
//
//		CPLFree( pszHDRFilename );
//        return( CE_Failure );
//	}
//
//	HFAClose(hHFA);
//	hHFA=NULL;
//
//	return( CE_None );
//}

//*****************************************************
//write section


/************************************************************************/
/*                      AIGProcess8bitIntBlockW()                       */
/*                                                                      */
/*      Process a block using ``08'' raw format.                        */
/************************************************************************/

static 
int AIGProcessRaw8BitBlockW( GInt32 * panData, int nMin,
                                int nBlockXSize, int nBlockYSize,  
                                GByte *pabyCur)

{
    int		i;
    
/* -------------------------------------------------------------------- */
/*      Collect raw data.                                               */
/* -------------------------------------------------------------------- */
    for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
    {
        *pabyCur = (GByte)(panData[i] - nMin);
        pabyCur++;
    }

    return nBlockXSize*nBlockYSize;
}


/************************************************************************/
/*                         AIGProcess16bitRawBlockW()                    */
/*                                                                      */
/*      Process a block using ``10'' (sixteen bit) raw format.          */
/************************************************************************/

static 
int AIGProcessRaw16BitBlockW( GInt32 * panData, int nMin,
                                 int nBlockXSize, int nBlockYSize, 
                                 GByte *pabyCur)

{
    int		i;

/* -------------------------------------------------------------------- */
/*      Collect raw data.                                               */
/* -------------------------------------------------------------------- */
    for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
    {
        GInt16	v = (GInt16)(panData[i]- nMin);
		v = CPL_MSBWORD16(v);

        *(pabyCur++) = ((GByte *) &v)[0];
        *(pabyCur++) = ((GByte *) &v)[1];
    }

    return nBlockXSize*nBlockYSize*2;
}

/************************************************************************/
/*                         AIGProcess32bitRawBlockW()                    */
/*                                                                      */
/*      Process a block using ``20'' (thirtytwo bit) raw format.        */
/************************************************************************/

static 
int AIGProcessRaw32BitBlockW( GInt32 * panData, GInt32 nMin, 
                                 int nBlockXSize, int nBlockYSize,
                                 GByte *pabyCur )

{
    int	  i;
    
/* -------------------------------------------------------------------- */
/*      Collect raw data.                                               */
/* -------------------------------------------------------------------- */
    for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
    {
        GInt32	v = panData[i]-nMin;
		v = CPL_MSBWORD32(v);

        *(pabyCur++) = ((GByte *) &v)[0];
        *(pabyCur++) = ((GByte *) &v)[1];
        *(pabyCur++) = ((GByte *) &v)[2];
        *(pabyCur++) = ((GByte *) &v)[3];
    }

    return nBlockXSize*nBlockYSize*4;
}

/************************************************************************/
/*                    AIGProcessRaw32bitFloatBlockW()                    */
/*                                                                      */
/*      Process a block using ``00'' (32 bit) raw format.               */
/************************************************************************/

static 
int AIGProcessRaw32BitFloatBlockW( float * pafData, int nBlockXSize, int nBlockYSize,
                                      GByte *pabyCur)

{
    int		i;
     
//* -------------------------------------------------------------------- *
//*      Collect raw data.                                               *
//* -------------------------------------------------------------------- *
    for( i = 0; i < nBlockXSize * nBlockYSize; i++ )
    {
		float	v = pafData[i];
		CPL_MSBPTR32(&v);

		*(pabyCur++) = ((GByte *) &v)[0];
        *(pabyCur++) = ((GByte *) &v)[1];
        *(pabyCur++) = ((GByte *) &v)[2];
        *(pabyCur++) = ((GByte *) &v)[3];
    }

    return nBlockXSize*nBlockYSize*4;
}




/************************************************************************/
/*                          AIGProcessBlock()                           */
/*                                                                      */
/*      Process a block using ``D7'', ``E0'' or ``DF'' compression.     */
/************************************************************************/

static 
int AIGProcessBlockW( GInt32 * panData, int nMin, int nMagic, 
                        int nBlockXSize, int nBlockYSize,  GByte *pabyCur)

{
    int		nTotPixels;
    int		i;
    int     nDataSize=0;

/* ==================================================================== */
/*     Process runs till we are done.                                   */
/* ==================================================================== */
    nTotPixels = nBlockXSize * nBlockYSize;
    i = 0;

    while( i < nTotPixels  )
    {
        int bNoData = panData[i]==ESRI_GRID_NO_DATA;
        int	nMarker = 0;
        int ii = i;
        
        while( nMarker<127 && ii < nTotPixels &&  (bNoData == (panData[ii]==ESRI_GRID_NO_DATA)) )
        {
            nMarker++;
            ii++;
        }

        if(bNoData)
            nMarker=256-nMarker;
        
        CPLAssert(nMarker>=1 && nMarker<256 );

        *(pabyCur++)  = (GByte) nMarker;
        nDataSize++;
        
/* -------------------------------------------------------------------- */
/*      Repeat data - four byte data block (0xE0)                       */
/* -------------------------------------------------------------------- */
        //if( nMagic == 0xE0 )
        //{
        //    GInt32	nValue;
        //    
        //    if( nMarker + nPixels > nTotPixels )
        //    {
        //        CPLError( CE_Failure, CPLE_AppDefined, 
        //                  "Run too long in AIGProcessBlock, needed %d values, got %d.", 
        //                  nTotPixels - nPixels, nMarker );
        //        return CE_Failure;
        //    }
        //
        //    if( nDataSize < 4 )
        //    {
        //        CPLError(CE_Failure, CPLE_AppDefined, "Block too small");
        //        return CE_Failure;
        //    }

        //    nValue = 0;
        //    memcpy( &nValue, pabyCur, 4 );
        //    pabyCur += 4;
        //    nDataSize -= 4;

        //    nValue = CPL_MSBWORD32( nValue );

        //    nValue += nMin;
        //    for( i = 0; i < nMarker; i++ )
        //        panData[nPixels++] = nValue;
        //}
        
/* -------------------------------------------------------------------- */
/*      Repeat data - two byte data block (0xF0)                        */
/* -------------------------------------------------------------------- */
        //else if( nMagic == 0xF0 )
        //{
        //    GInt32	nValue;
        //    
        //    if( nMarker + nPixels > nTotPixels )
        //    {
        //        CPLError( CE_Failure, CPLE_AppDefined, 
        //                  "Run too long in AIGProcessBlock, needed %d values, got %d.", 
        //                  nTotPixels - nPixels, nMarker );
        //        return CE_Failure;
        //    }

        //    if( nDataSize < 2 )
        //    {
        //        CPLError(CE_Failure, CPLE_AppDefined, "Block too small");
        //        return CE_Failure;
        //    }

        //    nValue = (pabyCur[0] * 256 + pabyCur[1]) + nMin;
        //    pabyCur += 2;
        //    nDataSize -= 2;

        //    for( i = 0; i < nMarker; i++ )
        //        panData[nPixels++] = nValue;
        //}
        
/* -------------------------------------------------------------------- */
/*      Repeat data - one byte data block (0xFC)                        */
/* -------------------------------------------------------------------- */
        //else if( nMagic == 0xFC || nMagic == 0xF8 )
        //{
        //    GInt32	nValue;

        //    if( nMarker + nPixels > nTotPixels )
        //    {
        //        CPLError( CE_Failure, CPLE_AppDefined, 
        //                  "Run too long in AIGProcessBlock, needed %d values, got %d.", 
        //                  nTotPixels - nPixels, nMarker );
        //        return CE_Failure;
        //    }
        //    
        //    if( nDataSize < 1 )
        //    {
        //        CPLError(CE_Failure, CPLE_AppDefined, "Block too small");
        //        return CE_Failure;
        //    }
        //
        //    nValue = *(pabyCur++) + nMin;
        //    nDataSize--;
        //    
        //    for( i = 0; i < nMarker; i++ )
        //        panData[nPixels++] = nValue;
        //}
        
/* -------------------------------------------------------------------- */
/*      Repeat data - no actual data, just assign minimum (0xDF)        */
/* -------------------------------------------------------------------- */
        //else if( nMagic == 0xDF && nMarker < 128 )
        //{
        //    if( nMarker + nPixels > nTotPixels )
        //    {
        //        CPLError( CE_Failure, CPLE_AppDefined, 
        //                  "Run too long in AIGProcessBlock, needed %d values, got %d.", 
        //                  nTotPixels - nPixels, nMarker );
        //        return CE_Failure;
        //    }
        //
        //    for( i = 0; i < nMarker; i++ )
        //        panData[nPixels++] = nMin;
        //}
        
/* -------------------------------------------------------------------- */
/*      Literal data (0xD7): 8bit values.                               */
/* -------------------------------------------------------------------- */
        if( nMagic == 0xD7 && nMarker < 128 )
        {
            while( nMarker > 0 )
            {
                *(pabyCur++)  = (GByte)(panData[i++] - nMin);
                nMarker--;
                nDataSize++;
            }
        }

/* -------------------------------------------------------------------- */
/*      Literal data (0xCF): 16 bit values.                             */
/* -------------------------------------------------------------------- */
        else if( nMagic == 0xCF && nMarker < 128 )
        {
            while( nMarker > 0 )
            {
                GInt16	v = (GInt16	)(panData[i++]-nMin);
				v = CPL_MSBWORD16(v);
				
				*(pabyCur++) = ((GByte *) &v)[0];
				*(pabyCur++) = ((GByte *) &v)[1];

                nMarker--;
                nDataSize += 2;
            }
        }

/* -------------------------------------------------------------------- */
/*      Nodata repeat                                                   */
/* -------------------------------------------------------------------- */
        else if( nMarker > 128 )
        {
            i+=(256-nMarker);
        }
        else
        {
            CPLAssert(FALSE);
        }

    }

    return nDataSize;
}
